/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 842 $ $LastChangedDate: 2002-09-01 18:56:58 +0200 (Sun, 01 Sep 2002) $
 *
 * Codegen - generate Ksi code from Gont Abstract Syntax Trees. 
 */

open Ast;
open Util;

open Cg_util;
open Cg_globals;
open Cg_expr;

<cg_ctx>option global_ctx_o = None;

<Ksi::id>list module_init = [];
<Ksi::id>list module_fini = [];

void init(Stdio::FILE ksi_file)
{
    Cg_util::ksi_sink = ksi_file;
    
    global_ctx_o = Some({cf_name = "?global_scope?",
			 cf_is_public = true,
			 cf_closure_type_name = "",
			 cf_ret_type_instanced = Abstract_type,
			 cf_references = ref_make(Nil),
			 cf_references_hash = Ptrhash::create(13),
			 parent = None,
			 parent_func = None,
			 assign_target = None});
    Ksi::output_header(ksi_sink);
    output_decl(Ksi::Input_file((top_location()).file));
    Cg_types::push_basic_types();
}

cg_ctx get_global_ctx()
{
    switch global_ctx_o {
    case Some(c):
	return c;
	
    case None:
	return ice("Codegen::get_global_ctx(): not inialized");
    }
}

void codegen_for_nl_decl(nl_decl d)
{
    cg_ctx c = get_global_ctx();
    
    switch d {
    case Func_decl(fn):
	let sym = get_binding(fn.fn_symbol) in {
	    push_fn_decl(c, ({name = sym.a.extern_name, 
			      is_public = has_property(sym, Is_exported),
			      takes_closure_parm = false,
			      lambda = fn.fn_lambda}));
	    // don't output new prototype for this function
	    set_property(sym, Is_prototyped);
	}
	
    case Section_decl(sec_type, l):
	let nam = genname(sec_type) in {
	    push_fn_decl(c, ({name = nam,
			      is_public = false, 
			      takes_closure_parm = false,
			      lambda = l}));

	    switch sec_type {
	    case "init":
		module_init = Cons(make_id(nam), module_init);
	    case "fini":
		module_fini = Cons(make_id(nam), module_fini);
	    case _:
		ice("Codegen::codegen_for_nl_decl(): " +
		    "bad section");
	    }
	}
    
    case Global_decl(v):
	// push proto actually outputs definition instead of declaration
	// for globals from current module
	push_proto(get_binding(v.var_symbol));
		    
    // symbols are prototyped when used
    case Extern_c(_, _):
    case Func_proto(_):
    case Tydef(_):

    // code for exceptions is generated during link
    case Exception_decl(_):

    // these don't generate any code
    case Top_level_open(_):
	skip;
    }
}

void codegen_for_decl(decl d)
{
    locate(codegen_for_nl_decl, d);
}

string push_function(<Ksi::expr>list exprs, string kind)
{
    void output(string name)
    {
	let body = Ksi::Bind(Ksi::Compound(exprs)) in
	let decl = Ksi::Func(List::cons1(Ksi::Public), Ksi::Void, make_id(name), 
			     Nil, body) in
	output_decl(decl);
    }

    switch exprs {
    case Nil:
	return "";
    case _:
	string name = Util::current_module + "$$" + kind + "$";
	output(name);
	return name;
    }
}

*(string, string, string) fini()
{
    <Ksi::expr>list call_lst(<Ksi::id>list names)
    {
	Ksi::expr call(Ksi::id fnc)
	{
	    return Ksi::Call(fnc, Nil);
	}
	
	return List::map(call, List::rev(names));
    }
	
    let ctors = push_function(get_global_inits(), "ctor") in
    let inits = push_function(call_lst(module_init), "init") in
    let finis = push_function(call_lst(module_fini), "fini") in
    let xgo_linked_id = make_id(Util::current_module + "$$xgo_linked$") in
    let xgo_linked = Ksi::Global([Ksi::Extern], null_type, xgo_linked_id) in
    let reference_xgo = Ksi::Modify(Ksi::Ref(xgo_linked_id), null_value) in
    let refxgo_fn = Ksi::Func([Ksi::Public], Ksi::Void,
			      make_id(Util::current_module + "$$reference_xgo$"),
			      Nil, Ksi::Bind(reference_xgo)) in
    {
	output_decl(xgo_linked);
	output_decl(refxgo_fn);

	Ksi::output_trailer(ksi_sink);

	return (ctors, inits, finis);
    }
}

// vim:sts=4:sw=4
