/*
 * Copyright (c) 2002 Michal Moskal <malekith@pld-linux.org>
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 1012 $ $LastChangedDate: 2002-09-21 20:16:14 +0200 (Sat, 21 Sep 2002) $
 *
 * Tyutil - Utilities for type manipulation.
 */

open Ast;
open Util;

/* Return string representation of [t], for example for error messages. 
   Safe to call on cyclic types.  */
string string_of_type(ty t)
{
    string string_of_field(<<'a>symbol>field f)
    {
	return string_of_type(f.fld_type) + " " + qualified_name(f.fld_symbol);
    }
    
    string record_content(<<<'a>symbol>field>list lst)
    {
	return String::concat("; ", List::map(string_of_field, lst));
    }
    
    switch t {
    case Int:	    return "int";
    case Bool:	    return "bool";
    case Void:	    return "void";
    case String:    return "string";
    case Float:	    return "float";
    case Char:	    return "char";
    case Exn:	    return "exn";
    case Abstract_type:	return "abstract-type";
    case Tuple(lst):
	let args = List::map(string_of_type, lst) in
	return "*(" + String::concat(", ", args) + ")";
    case Func(ft):
	let args = List::map(string_of_type, List::map(fst, ft.parms)) in
	let tyvars = List::map(fun (s) ("'" + s), ft.tyvars) in
	return "*(<" + String::concat(", ", tyvars)  + "> " +
	       string_of_type(ft.ret_type) + 
	       " (" + String::concat(", ", args) + "))";
    case Tyvar(n):
	return "'" + n;
	
    case X_tyvar(x):
	let hd = "'" + x.name + " [" in 
	switch x.t {
	case Some(tt):
	    if (x.flag) {
		x.flag = false;
		switch tt {
		case X_tyvar(xc):
		    return hd + xc.name + " (cycle)]";
		case _:
		    return hd + "cycle]";
		}
	    }
	    x.flag = true;
	    _ st = string_of_type(tt);
	    x.flag = false;
	    return hd + st + "]";
	case None:
	    return hd + "?]";
	}

    case Struct(lst):
	return "struct {" + record_content(lst) + "}";

    case Opt_struct(lst):
	return "opt_struct {" + record_content(lst) + "}";

    case Union(lst):
	return "union {" + record_content(lst) + "}";
    
    case Named_type(nt):
	let args = List::map(string_of_type, nt.parms) in
	let qn = qualified_name(nt.tydef) in
	switch args {
	case Nil:
	    return qn;
	case Cons(_, _):	
	    return "<" + String::concat(", ", args) + ">" + qn;
	}
    }
}

/* Return [t] with all levels of X_tyvar(...) indirection stripped off.
   It is safe to call on cyclic type in which case it will report error
   message and return Abstract_type.  */
ty normalize_xt(ty t)
{
    switch t {
    case X_tyvar(x):
	if (x.flag) {
	    error("type " + string_of_type(t) + " is cyclic");
	    return Abstract_type;
	}
	
	switch x.t {
	case Some(tt):
	    x.flag = true;
	    t = normalize_xt(tt);
	    x.flag = false;
	
	case None:
	    skip;
	}
	
    case _:
	skip;
    }
    
    return t;
}

/* Strip all X_tyvar() stuff until some real type.  ICE if None if found.  
   To be used by passes after typecheck, which can generally assume there
   are no weak types in what they operate on.  */
ty strip_xt(ty t)
{
    switch normalize_xt(t) {
    case X_tyvar(_):
	return ice("Type::strip_xt(): weak type variable");
    case x:
	return x;
    }
}

/* Substitute X_tyvar(...) in type with their content.  Often good for error 
   messages, but not always.  */
ty clean_type(ty t)
{
    return type_map(normalize_xt, t);
}

/* Substitute fresh type variables in [t] for type variables listed in [ns].
   Return list of variables substituted and substituted type.  Probably there
   is no need to use this beside Type and Tyutil modules.  */
*(<xtyvar>list, ty) subst_xtyvars(ty t, <id>list ns)
{
    <xtyvar>Hashtbl::t vars = Hashtbl::create(13);
    <xtyvar>list ret = Nil;

    List::iter(fun (s) {
		    let x = genxt("_s_" + s) in {
			ret = Cons(x, ret);
			if (Hashtbl::mem(vars, s))
			    ice("Type::subst_xtyvars(): type variable "
				"specified twice: " + s);
			Hashtbl::add(vars, s, x);
		    }
	       }, ns);
	
    ty f(ty x)
    {
	switch x {
	case Tyvar(a):
	    try {
		return X_tyvar(Hashtbl::find(vars, a));
	    } with {
	    case Not_found:
		return x;
	    }
	    
	case X_tyvar(xt):
	    switch xt.t {
	    case Some(s):
		return f(s);
	    case None:
		return x;
	    }
	
	case _:
	    return x;
	}
    }

    return (List::rev(ret), type_map(f, t));
}

/* For each pair [(name, type)] in [assoc] substitute all occurrences of
   [Tyvar(name)] with [type] in [t].  Return substituted [t].  Probably there
   is no need to use this beside Type and Tyutil modules.  */
ty subst_vars(ty t, <*(id, ty)>list assoc)
{
    let (xts, new_t) = subst_xtyvars(t, List::map(fst, assoc)) in {
	void f(xtyvar x, ty s) { x.t = Some(s); }
	List::iter2(f, xts, List::map(snd, assoc));
	return new_t;
    }
}

/* Strip one level of Named_type(...) indirection.  Type arguments are 
   substituted as necessary.  Will ICE on weak types.  */
ty type_instance_one(ty t)
{
    switch strip_xt(t) {
    case Named_type(nt):
	let ts = get_binding(nt.tydef) in
	let assoc = List::map2(fun (tt, n) ((tt, n)), 
			       ts.a.tysym_tyvars, nt.parms) in
	return subst_vars(ts.stype, assoc);
    
    case x:
	return x;
    }
}

/* Test whether [t] is base type, i.e. is not just an alias for some other 
   type.  Will ICE on weak types.  */
bool is_base_type(ty t)
{
    switch strip_xt(t) {
    case Named_type(nt):
	let ts = get_binding(nt.tydef) in
	switch ts.stype {
	case Struct(_):	    return true;
	case Opt_struct(_):  return true;
	case Union(_):	    return true;
	case Abstract_type:  return true;
	case Named_type(_):  return false;
	case _:		    return false;
	}
	
    case _:
	return true;
    }
}

/* Test whether [t] is fully instanced type, i.e. is not Named_type(...).  
   Will ICE on weak types.  */
bool is_type_instance(ty t)
{
    switch strip_xt(t) {
    case Named_type(_):
	return false;
    
    case _:
	return true;
    }
}

/* Return real content of type, i.e. with all Named_type(...) stripped off.
   It will return Struct(...), Opt_struct(...) and Union(...) as appropriate.
   Won't ever return Named_type(...) or X_tyvar(...).  Will ICE on weak types.  */
ty type_instance(ty t)
{
    t = strip_xt(t);

    while (!is_type_instance(t))
	t = type_instance_one(t);

    return t;
}

/* Return copy of [tt] with all type aliases expanded to their real meaning.
   However Named_types(...) for structures and unions are not expanded.
   This functions just expands toplevel [t], not its content.  See 
   canonical_type() for recursive version. Will ICE on weak types.  */
ty base_type(ty t)
{
    t = strip_xt(t);

    while (!is_base_type(t))
	t = type_instance_one(t);
    
    return t;
}

/* Same as base_type() except it traverses entire type tree expanding
   Named_types(...).  */
ty canonical_type(ty tt)
{
    ty f(ty t)
    {
	switch t {
	case Named_type(_):
	    return base_type(t);
	case X_tyvar(xt):    
	    switch xt.t {
	    case Some(s):
		return canonical_type(s);
	    case None:
		return ice("Codegen::rec_base_type(): weak tyvar");
	    }
	case _:
	    return t;
	}
    }

    return type_map(f, tt);
}

/* Return type of [e].  Can only be called after expression has been type 
   checked.  */
ty type_of(expr e)
{
	switch e.a.expr_type {
	case Some(t):
		return t;
	case None:
		return ice("Util::type_of(): untyped expression");
	}
}

/* Return fully expanded type of [e].  Can only be called after expression
   has been type checked.  */
ty instanced_type_of(expr e)
{
    return type_instance(type_of(e));
}

/* Return [true] iff [t] is type of function returning Void.  */
bool is_void_func_type(ty t)
{
    let ft = get_ft_record(type_instance(t)) in
    switch type_instance(ft.ret_type) {
    case Void:
	return true;
    case _:
	return false;
    }
}

/* Check if [e] is valid L-value, i.e. can be used as target of assignment 
   expression.  */
bool is_lvalue(expr e)
{
    switch e.a.raw {
    case Ref(rr):
	return has_property(get_binding(rr.ref_symbol), Is_lvalue);

    case Dot_expr(_):
	return true;
	
    case Tuple_expr(l):
	return !List::exists(fun bool (expr ex) (!is_lvalue(ex)), l);
	
    case _:
	return false;
    }
}

/* Lookup union of given one of labels contained in it.  */
ty_symbol union_of(var_symbol label)
{
    switch label.a.parent_union {
    case Some(ts):
	return ts;
    case None:
	return ice("Type::union_of(): not an union");
    }
}

/* Given a structure type, lookup named field in it.  Print error message
   and raise Choke if not found.  */
<sf_symbol>field find_field_in_struct(ty t, <sf_symbol>bindable name)
{
    id sn = simple_name(name);
   
    bool is_our_fld(<<'a>symbol>field f)
    {
	return simple_name(f.fld_symbol) == sn;
    }
    
    switch type_instance(t) {
    case Struct(content):
    case Opt_struct(content):
	try {
	    return List::find(is_our_fld, content);
	} with {
	case Not_found:
	    error("`" + sn + "' is not part of " + string_of_type(t));
	    raise Choke;
	}
	
    case _:
	return ice("Type::type_of_field_in(): not record type");
    }
}

/* Given name of field. lookup its type in given union/struct type.  */
ty field_type_in(ty t, <<'a>symbol>bindable name)
{
    id sn = simple_name(name);
    
    bool is_our_fld(<<'b>symbol>field f)
    {
	return simple_name(f.fld_symbol) == sn;
    }
    
    switch type_instance(t) {
    case Union(content):
    case Struct(content):
    case Opt_struct(content):
	try {
	    return (List::find(is_our_fld, content)).fld_type;
	} with {
	case Not_found:
	    error("`" + sn + "' is not part of " + string_of_type(t));
	    raise Choke;
	}
	
    case _:
	return ice("Type::type_of_field_in(): not record type");
    }
}

/* Check whatever given expression is constant, i.e. can be evaluated at
   compile-time.  This is kinda conservative.  */
bool is_constant_expr(expr e)
{
    switch e.a.raw {
    case Simple_binop(_, e1, e2):
	return is_constant_expr(e1) && is_constant_expr(e2);
	
    case Simple_unop(_, e1):
	return is_constant_expr(e1);
    
    case Ref(_):
    case Call(_, _):
    case Assign(_, _):
    case Assign_target:
	return false;

    case Tuple_expr(lst):
	return !List::exists(fun (x) (!is_constant_expr(x)), lst);

    case Union_expr(u):
	return is_constant_expr(u.nx_expr);
    
    case Lambda(_):
    case Struct_expr(_):
    case Struct_copy(_):
    case Dot_expr(_):
	return false;
	
    case Bool_const(_):
    case Int_const(_):
    case Float_const(_):
    case String_const(_):
    case Char_const(_):
    case Void_const:
    case Null_const:
	return true;
    }
}

/* Check whatever given symbol is function defined in global scope, which
   means it has different calling convention -- no closure.  */
bool is_global_function(var_symbol s)
{
    switch type_instance(s.stype) {
    case Func(_) if s.a.fnc_level <= 0 && !has_property(s, Is_global_var):
	return true;
    case _:
	return false;
    }
}

/* If [t] is an function type, return its [ft_record].  Otherwise ICE.  */
ft_record get_ft_record(ty t)
{
	switch t {
	case Func(ft):
		return ft;
		
	case _:
		return ice("Util::get_ft_record(): non-func type");
	}
}

/* Return string being flat terse representation of [t], using 
   [[A-Za-z_0-9:<>]] characters only.  It's used for debugging information.  */
string mangle_type(ty t)
{
    string mangle_symbol(<'a>symbol s)
    {
	return s.module + "::" + s.name;
    }
    
    string mangle_field(<<'a>symbol>field f)
    {
	    return mangle_type(f.fld_type) + 
		   mangle_symbol(get_binding(f.fld_symbol));
    }
    
    string record_content(<<<'a>symbol>field>list lst)
    {
	    return String::concat("><", List::map(mangle_field, lst));
    }

    string mangle_list(<ty>list l)
    {
    	return String::concat("", List::map(mangle_type, l));
    }

    switch t {
    case Int:		return "i";
    case Bool:		return "b";
    case Void:		return "v";
    case String:	return "s";
    case Float:		return "f";
    case Char:		return "c";
    case Exn:		return "e";
    case Abstract_type:	return "a";
    case Tuple(lst):
	return "T<" + mangle_list(lst) + ">";
	
    case Func(ft):
	return "F<" + mangle_type(ft.ret_type) + 
		mangle_list(List::map(fst, ft.parms)) + ">";
		   
    case Tyvar(n):
	return "V<" + n + ">";
	    
    case X_tyvar(x):
    	switch x.t {
	case Some(tt):
		return mangle_type(tt);
	case None:
		return "X<" + x.name + ">";
	}

    case Struct(lst):
	    return "S<<" + record_content(lst) + ">>";

    case Opt_struct(lst):
	    return "O<<" + record_content(lst) + ">>";

    case Union(lst):
	    return "U<<" + record_content(lst) + ">>";
    
    case Named_type(nt):
	return "N<<" + mangle_list(nt.parms) + "><" + 
	       mangle_symbol(get_binding(nt.tydef)) + ">>";
    }
}

/* Generate new weak type variable, name of which will start with [name].  */
xtyvar genxt(string name)
{
    return {name = genname(name), t = None, flag = false};
}

/* Check if there are none weak type variables in [t].  If they are, issue
   error message and return [true], otherwise return [false].  */
bool check_weak(ty t)
{
    bool has_weak = false;
    
    void f(ty tt)
    {
	switch tt {
	case X_tyvar(x):
	    switch x.t {
	    case None:
		has_weak = true;
	    case Some(_):
		skip;
	    }
	case Tyvar("$dummy"):
	    has_weak = true;
	case _:
	    skip;
	}
    }

    type_iter(f, t);
    
    if (has_weak)
	error("weak type variables not allowed here (type is " + 
		  string_of_type(t) + ")");

    return has_weak;
}

/* Return conservative estimation whether [ex] might have any side effects.
   Allocation is not counted as a side-effect.  */
bool has_side_effects(expr ex)
{
    switch ex.a.raw {
    case Assign(_, _):
	return true;
	
    case Assign_target:
	return ice("Cg_util::has_side_effects(): assign target");
	
    case Simple_binop(_, e1, e2):
	return has_side_effects(e1) || has_side_effects(e2);
	
    case Simple_unop(op, e):
	switch op {
	case Not:
	case Bit_not:
	case Negate:	return has_side_effects(e);
	case Pre_dec:	return true;
	case Pre_inc:	return true;
	case Post_dec:	return true;
	case Post_inc:	return true;
	}
    
    case Ref(_):
	return false;

    case Call(_, _):
	return true;

    case Tuple_expr(lst):
	return List::exists(has_side_effects, lst);

    case Union_expr(u):
	return has_side_effects(u.nx_expr);
    
    // we lay a bit here, as generating lambda is lots of work
    // in case of lack of closure
    case Lambda(_):
	return true;

    case Struct_expr(lst, _):
	return List::exists(fun (s) (has_side_effects(s.sx_expr)), lst);
    
    case Struct_copy(e, lst, _):
	return has_side_effects(e) ||
	       List::exists(fun (s) (has_side_effects(s.sx_expr)), lst);

    case Dot_expr(de, _):
	return has_side_effects(de.sx_expr);
    
    case Char_const(_):
    case Bool_const(_):
    case Int_const(_):
    case Float_const(x):
    case String_const(x): 
    case Void_const:
    case Null_const:
	return false;
    }
}

/* Return [true] iff [t] contains any type variables.  */
bool is_poly(ty t)
{
    bool is = false;
    
    void poly_check(ty t)
    {
	switch t {
	case Tyvar(_):
	    is = true;
	case _:
	    skip;
	}
    }

    type_iter(poly_check, t);

    return is;
}

// vim:sts=4:sw=4
