/* 
 * Copyright (c) 2001, 2002 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Rev: 1048 $ $LastChangedDate: 2002-10-26 23:19:49 +0200 (Sat, 26 Oct 2002) $
 */

#ifndef GONT_RUNTIME_H
#define GONT_RUNTIME_H 1

#include <stddef.h>

/*** Basic types. ***/

/* Integer of the same size, as pointer. */
#define gont_int unsigned long

#define GONT_WORD_SIZE (sizeof(gont_int) * 8)

/*** Memory management. ***/

/* All gont_xmalloc* routines prints error message and aborts if there
 * is not enough memory to satisfy request. */

/* Initialize allocator. */
void gont_init_memory_allocator();

/* Allocate given amount of memory. */
void *gont_xmalloc(gont_int s);

/* Allocate given amount of memory and clear it. */
void *gont_xmalloc0(gont_int s);

/* Allocate given amount of memory, one cannot put pointers into this memory 
 * then (for example this is used for strings). */
void *gont_xmalloc_atomic(gont_int s);

/* As above, and clear it. */
void *gont_xmalloc0_atomic(gont_int s);

void gont_register_finalizer(void *obj, void (*fin)(void *obj, void *data), 
			     void *data);

#define gont_free(x) do {} while (0)


/*** Initialization/finalization. ***/
extern int gont_argc;
extern char **gont_argv;
extern char **gont_environ;
extern int gont_exit_code;

void gont_init_runtime(void);
void gont_set_argv(int argc, char **argv);
void gont_set_environ(char **envp);
void gont_register_exit_routine(void (*fnc)(void*), void *data);
void gont_run_exit_routines(void);
void gont_exit(void);

/* Print error message and abort. */
void gont_oops(const char *msg, ...) __attribute__ ((__noreturn__));

/*** Exceptions. ***/
struct gont_try_block;		/* Hidden. */

/* Value returned from this function can be used as __builtin_setjmp 
 * buffer. */
struct gont_try_block *gont_gont_try_block(void);

/* Pop value from try-block chain. */
void *gont_pop_raised_value(void);

/* Just pop one block, throw away the value. */
void gont_pop_try_block(void);

/* Raise exception. */
void gont_raise(void *v) __attribute__ ((__noreturn__));

void gont_fail_with(const char *s) __attribute__ ((__noreturn__));
void gont_invalid_argument(const char *s) __attribute__ ((__noreturn__));
void gont_sys_error(gont_int err) __attribute__ ((__noreturn__));
void gont_invalid_mbseq(void) __attribute__ ((__noreturn__));
void gont_incomplete_mbseq(void) __attribute__ ((__noreturn__));
void gont_invalid_unicode_character(void) __attribute__ ((__noreturn__));
void gont_character_not_supported_by_encoding(void) __attribute__ ((__noreturn__));


/*** Complex Gont datatypes. ***/

/* Lambda expressions. */
struct gont_lambda_expr {
	void *(*fnc)(void *, ...);	/* Pointer to function. */
	void *closure;			/* Pointer to its closure. */
};

/* Make new lambda expression. */
struct gont_lambda_expr *gont_make_lambda(void *fnc, void *closure);


/* Strings. */
struct gont_string;

unsigned char *gont_string_to_utf8(unsigned char *buf_hint, gont_int buf_size,
				   struct gont_string *s, gont_int *sizep);
gont_int *gont_string_to_utf32(struct gont_string *s);
struct gont_string *gont_string_of_utf8(const unsigned char *ptr, gont_int bytes);
struct gont_string *gont_string_of_utf32(const gont_int *ptr, gont_int chars);

struct gont_charset;
void gont_free_charset(struct gont_charset *ch);
struct gont_charset *gont_make_charset(const char *ch, gont_int do_approx,
				       struct gont_lambda_expr *approx_fnc);
unsigned char *gont_string_to_cstring(unsigned char *buf_hint, 
				      gont_int hint_size,
				      struct gont_string *s,
				      gont_int *sizep,
				      struct gont_charset *ch);
struct gont_string *gont_string_of_cstring(const unsigned char *str, gont_int sz,
					   struct gont_charset *ch);
extern struct gont_charset *gont_default_charset;
extern struct gont_charset *gont_default_charset_with_approx;

struct gont_string *gont_string_utf8(const char *str);
struct gont_string *gont_string_locale(const char *str);

int gont_append_approx(struct gont_lambda_expr *approx_fnc, 
		       char **in_buf, size_t *in_left,
		       char **out_buf, size_t *out_left);
void gont_charset_set_approx(struct gont_charset *ch, gont_int do_approx,
			     struct gont_lambda_expr *approx_fnc);

gont_int gont_streq(struct gont_string *s1, struct gont_string *s2);
gont_int gont_strcmp(struct gont_string *s1, struct gont_string *s2);
struct gont_string *gont_strjoin(struct gont_string *s1, 
			         struct gont_string *s2);
/* Result is always in [0, 0x7fffffff]. */
gont_int gont_hash_string(struct gont_string *s);

/* Identity function. */
gont_int gont_id(gont_int x);

/* Floats. */
double *gont_make_float(double v);

#define GONT_STRING_TO_CSTRING(s, ch) \
	char cstring_buf_dont_touch[1024];	\
	unsigned char *cstring;			\
	gont_int cstring_len;			\
	cstring = gont_string_to_cstring(cstring_buf_dont_touch, 	\
					 sizeof(cstring_buf_dont_touch),\
					 s, &cstring_len, ch);

#define GONT_STRING_TO_CSTRING_N(s, ch) 	\
	char cstring_buf_dont_touch ## s[1024];	\
	gont_int c_ ## s ## _len;		\
	unsigned char *c_ ## s = 		\
		gont_string_to_cstring(cstring_buf_dont_touch ## s, 		\
				       sizeof(cstring_buf_dont_touch ## s),	\
				       s, &c_ ## s ## _len, ch);


/* Unions. */
struct gont_union {
	gont_int field_no;
	void *value;
};

struct gont_union *gont_make_union(gont_int field_no, void *value);
void **gont_make_tuple(gont_int n, ...);

const char *gont_get_locale_charset(void);

void *gont_memchr(void *haystack, gont_int haystack_len, gont_int ch);
void *gont_memrchr(void *haystack, gont_int haystack_len, gont_int ch);
void *gont_memmem(void *haystack, gont_int haystack_len,
		  void *needle, gont_int needle_len);
void *gont_memrmem(void *haystack, gont_int haystack_len,
		   void *needle, gont_int needle_len);

struct gont_string *gont_strjoin(struct gont_string *s1, 
			         struct gont_string *s2);
gont_int gont_streq(struct gont_string *s1, struct gont_string *s2);
gont_int gont_strcmp(struct gont_string *s1, struct gont_string *s2);
gont_int gont_string_hash(struct gont_string *s);
gont_int gont_string_length(struct gont_string *s);
struct gont_string *gont_string_sub(struct gont_string *s, gont_int pos, 
				    gont_int len);
gont_int gont_string_get(struct gont_string *s, gont_int idx);
struct gont_string *gont_string_make(gont_int n, gont_int c);
gont_int gont_string_index_from(struct gont_string *s, gont_int pos, gont_int c);
gont_int gont_string_rindex_from(struct gont_string *s, gont_int pos, gont_int c);
gont_int gont_string_index_str_from(struct gont_string *s, gont_int pos, 
				    struct gont_string *c);

#endif /* GONT_RUNTIME_H */
